#include "spp_conversation.h"

/***********************Global vars*********************/
StateRecord*		states;
StateRecord**		state_hash;
StateRecord*		last;
unsigned long		last_session_id;

PacketHolder*		caches;
PacketHolder*		last_cache;

#ifdef DEBUG
unsigned int MaxDepth;
#endif

/***********************Function Declaration*************/
int init_state();
int assign_session_id(Packet* p);

/****************************************
*  Register the preprocessor
****************************************/
void SetupConv(){
	RegisterPreprocessor("conversation", ConvInit);
	
#ifdef DEBUG
	printf("Preprocessor: Registering Session\n");
#endif
}

/****************************************
*  Initialize everything
****************************************/
void ConvInit(u_char* args){
	init_state();

	AddFuncToPreprocList(ConvFunc);
}

/****************************************
*  Called for every packet
****************************************/
void ConvFunc(Packet* p){
	if (p->iph)
	assign_session_id(p);	
}

/***************************************
* create all the structs and init vars
***************************************/
int init_state(){
	int i;
	
#ifdef DEBUG
	printf("Initializing State Engine\n");
#endif
	states=(StateRecord*)malloc(sizeof(StateRecord) * NUM_STATES);
	bzero(states, sizeof(StateRecord) * NUM_STATES);
	state_hash=(StateRecord**)malloc(sizeof(StateRecord*) * STATE_HASH_SIZE);
	bzero(state_hash, sizeof(StateRecord*)*STATE_HASH_SIZE);
	
	/*chain together the ring*/
	for (i=0; i<NUM_STATES; i++){
		states[i].next = &states[i+1];
	}
	states[NUM_STATES-1].next = &states[0];
	
	last=&states[0];
	last_session_id=0;
	
	/*alloc the cache stuff*/
	caches=(PacketHolder*)calloc(sizeof(PacketHolder)*NUM_CACHE, sizeof(char));
	/*chain together the ring*/
	for (i=0;i<NUM_CACHE-1;i++){
		caches[i].next=&caches[i+1];
	}
	caches[NUM_CACHE-1].next = &caches[0];
	last_cache=&caches[NUM_CACHE-1];

#ifdef DEBUG	
	MaxDepth=0;
#endif	
	
	return TRUE;
}

/****************************************************
* hash this down for quick lookups
****************************************************/
unsigned short HashPacket(Packet* p){
	unsigned long	hash;

#ifdef DEBUG
	//printf("Hashing Packet\n");
#endif	
	hash=*(unsigned long*)&p->iph->ip_src;
	hash=hash ^ *(unsigned long*)&p->iph->ip_dst;
	hash = hash ^ p->sp;
	hash = hash ^ p->dp;
	hash = hash + p->iph->ip_proto;

	return (hash/65535) + ((hash * 65535)/65535);
}

/***********************************************
* get a free state from the ring
*
* TODO: delete the hash sanely
***********************************************/
StateRecord* get_free_state(unsigned long cur_time){
	StateRecord*	this;
	StateRecord*	this2;
	int				count;
	
#ifdef DEBUG	
	printf("Getting free state\n");
#endif	
	this=last->next;
	count=0;
	
	while (TRUE){
		if (this->last_time==0){
			/*never been allocated before*/
			last=this;
			return this;
		}
		if ( (cur_time - this->last_time) > STATE_TIMEOUT ){
			/*re-used record*/
#ifdef DEBUG			
			printf("Reusing Record\n");
#endif			
			/*unlink this one*/
			if (*this->hash_prev == this){
				*(this->hash_prev) = this->hash_next;
			}else{
				printf("FIXME: prev->next was not this\n");
			}	
			/*take care of the stray pointers*/
			this->hash_next=NULL;
			this->hash_prev=NULL;
			this->cache=NULL;
			last=this;
			return this;
		}
#ifdef DEBUG		
		printf("This state still in use\n");
#endif		
		if (count>MAX_STATE_WALK){
#ifdef DEBUG
			printf("We've looked for a state too long, expire the oldest\n");
#endif		
			/*unlink this one*/
			this=last->next;
			if (*this->hash_prev == this){
				*(this->hash_prev) = this->hash_next;
			}else{
				printf("FIXME: prev->next was not this\n");
			}	
			/*take care of the stray pointers*/
			this->hash_next=NULL;
			this->hash_prev=NULL;
			this->cache=NULL;
			last=this;
			return this;
		}
		this=this->next;
		count++;
	}
}

/***********************************************
* Fills the state record with info
***********************************************/
void fill_state(StateRecord* s, Packet* p){
	if ( *((unsigned int*)&p->iph->ip_dst) > *((unsigned int*)&p->iph->ip_src) ){
		memcpy(&s->SIP, &p->iph->ip_src, sizeof(struct in_addr));
		memcpy(&s->DIP, &p->iph->ip_dst, sizeof(struct in_addr));
		s->protocol=p->iph->ip_proto;
		s->sp = p->sp;
		s->dp = p->dp;
	}else{
		memcpy(&s->SIP, &p->iph->ip_dst, sizeof(struct in_addr));
		memcpy(&s->DIP, &p->iph->ip_src, sizeof(struct in_addr));
		s->protocol=p->iph->ip_proto;
		s->sp = p->dp;
		s->dp = p->sp;
	}
}

/***********************************************
* Find a session from the hash.
* returns a pointer to that session.
* allocates a new session if not found
***********************************************/
StateRecord* find_session(Packet* p){
	unsigned short 	hash;
	StateRecord*	this;
#ifdef DEBUG
	unsigned int	Depth;
	Depth=0;
#endif	

	hash=HashPacket(p);
	
	if (!state_hash[hash]){
		state_hash[hash]=get_free_state(p->pkth->ts.tv_sec);
		this=state_hash[hash];
#ifdef DEBUG		
		printf("First in Hash\n");
#endif		
		fill_state(this, p);
		this->hash_prev=&state_hash[hash];
		p->state=this;
		this->session_id=++last_session_id;
		return this;
	}
	
	this = state_hash[hash];
	while (TRUE){
		/*switch for directional stuff*/
		if ( *((unsigned int*)&p->iph->ip_dst) > *((unsigned int*)&p->iph->ip_src) ){
#ifdef DEBUG		
			printf("Comparing %s:%u to ", inet_ntoa(this->SIP), this->sp);
			printf("%s:%u\n", inet_ntoa(p->iph->ip_src), p->sp);
#endif			
			if (
			  (memcmp(&this->SIP, &p->iph->ip_src, sizeof(struct in_addr))==0) &&
			  (memcmp(&this->DIP, &p->iph->ip_dst, sizeof(struct in_addr))==0) &&
			  (this->sp == p->sp) &&
			  (this->dp == p->dp) &&
			  (this->protocol == p->iph->ip_proto)
			){
#ifdef DEBUG			
				printf("Found state forward\n");
#endif				
				p->state=this;
				return this;
			}
		}else{
#ifdef DEBUG		
			printf("Comparing %s:%u to ", inet_ntoa(this->SIP), this->sp);
			printf("%s:%u\n", inet_ntoa(p->iph->ip_dst), p->dp);
#endif			
			if (
			  (memcmp(&this->SIP, &p->iph->ip_dst, sizeof(struct in_addr))==0) &&
			  (memcmp(&this->DIP, &p->iph->ip_src, sizeof(struct in_addr))==0) &&
			  (this->sp == p->dp) &&
			  (this->dp == p->sp) &&
			  (this->protocol == p->iph->ip_proto)
			){
#ifdef DEBUG			
				printf("Found state backward\n");
#endif				
				p->state=this;
				return this;
			}
		}
		
		if (this->hash_next){
#ifdef DEBUG					
			printf("Next in chain\n");
			Depth++;
#endif			
			this = this->hash_next;
		}else{
#ifdef DEBUG		
			printf("Hit end of chain\n");
#endif			
			this->hash_next = get_free_state(p->pkth->ts.tv_sec);
			this->hash_next->hash_prev = &this->hash_next;
			fill_state(this->hash_next, p);
			p->state=this->hash_next;
			this->hash_next->session_id=++last_session_id;
#ifdef DEBUG
			if (Depth > MaxDepth){
				MaxDepth=Depth;
				printf("++++++++++++++++++++++++++++++++++++++++++++++++++++++New Max %u\n",MaxDepth);
			}
#endif				
			return this->hash_next;
		}	
	}
		
}

/***********************************************
* Process the packet 
***********************************************/
int assign_session_id(Packet* p){
	StateRecord*	this;

#ifdef DEBUG	
	printf("____________________________________________\n");
	printf("%s:%u->",inet_ntoa(p->iph->ip_src), p->sp);
	printf("%s:%u\n",inet_ntoa(p->iph->ip_dst), p->dp);
#endif	

	this = find_session(p);
	this->last_time = p->pkth->ts.tv_sec;
#ifdef DEBUG	
	printf("Session id: %lu\n",((StateRecord*)p->state)->session_id);
#endif	
	return TRUE;
}

/**********************************************
* Store some data for future reference
**********************************************/
int	CachePacket(StateRecord* s, Packet*	p){

#ifdef DEBUG
	printf("Caching Packet\n");
#endif	
	
	last_cache = last_cache->next;	
	
	if (!last_cache){
		printf("FIXME: last_cache was NULL!\n");
		last_cache=&caches[0];
		return FALSE;
	}
	
	/*erase the entry if it's in use*/
	if (last_cache->cache_prev){
		*(last_cache->cache_prev)=last_cache->cache_next;
		bzero(&last_cache->payload[0], PACKET_SIZE);
		last_cache->cache_next=NULL;
		last_cache->cache_prev=NULL;
	}
	
	/*fill in the entry*/
	if (p->dsize < PACKET_SIZE){
		memcpy(last_cache->payload + (PACKET_SIZE-p->dsize), p->data, p->dsize);
		last_cache->len=p->dsize;
	}else{
		/*copy last 600 bytes into buffer*/
		memcpy(last_cache->payload, p->data + (p->dsize-PACKET_SIZE) , PACKET_SIZE);
		last_cache->len=PACKET_SIZE;
	}
	last_cache->seq=ntohl(p->tcph->th_seq) + p->dsize;		
	
	/*link it on the chain*/
	if (!s->cache){
		s->cache=last_cache;
		s->cache->cache_prev=&s->cache;
		return TRUE;
	}
	
	last_cache->cache_next=s->cache;
	s->cache->cache_prev=&(last_cache->cache_next);
	last_cache->cache_prev=NULL;
	s->cache=last_cache;
	
	return TRUE;	
}

/**********************************************
* Get some data back
*
* seq is the sequence number of the next packet
* not the packet being searched for.
**********************************************/
PacketHolder* GetCachePacket(StateRecord* s, unsigned int seq){
	PacketHolder*	this;
	int				count;

#ifdef DEBUG	
	printf("Searching for seq %u\n",seq);
#endif	
	
	if (!s->cache){
		return NULL;
	}
	
	this=s->cache;
	count=0;
	while (this){
		if (this->seq == seq) return this;
		this=this->cache_next;
		count++;
		if (count>MAX_CACHE_WALK){
#ifdef DEBUG
			printf("We've searched too long, bail\n");
#endif		
			return NULL;
		}
	}
	
	return NULL;
}
